# Licensed to the Apache Software Foundation (ASF) under one or more§
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import random
import socket
import string
import os.path
import platform
import unittest
import warnings
from itertools import chain

import pytest

import libcloud.utils.files
from libcloud.utils.py3 import StringIO, b, bchr, urlquote, hexadigits
from libcloud.utils.misc import get_driver, set_driver, get_secure_random_string
from libcloud.common.types import LibcloudError
from libcloud.compute.types import Provider
from libcloud.utils.publickey import get_pubkey_ssh2_fingerprint, get_pubkey_openssh_fingerprint
from libcloud.utils.decorators import wrap_non_libcloud_exceptions
from libcloud.utils.networking import (
    is_public_subnet,
    is_private_subnet,
    join_ipv4_segments,
    is_valid_ip_address,
    increment_ipv4_segments,
)
from libcloud.compute.providers import DRIVERS
from libcloud.compute.drivers.dummy import DummyNodeDriver
from libcloud.storage.drivers.dummy import DummyIterator

# In Python > 2.7 DeprecationWarnings are disabled by default
warnings.simplefilter("default")


WARNINGS_BUFFER = []

from io import FileIO as file


def show_warning(msg, cat, fname, lno, file=None, line=None):
    WARNINGS_BUFFER.append((msg, cat, fname, lno))


original_func = warnings.showwarning


class TestUtils(unittest.TestCase):
    def setUp(self):
        global WARNINGS_BUFFER
        WARNINGS_BUFFER = []

    def tearDown(self):
        global WARNINGS_BUFFER
        WARNINGS_BUFFER = []
        warnings.showwarning = original_func

    @unittest.skipIf(platform.system().lower() == "windows", "Unsupported on Windows")
    def test_guess_file_mime_type(self):
        file_path = os.path.abspath(__file__)
        mimetype, encoding = libcloud.utils.files.guess_file_mime_type(file_path=file_path)

        self.assertTrue(mimetype.find("python") != -1)

    def test_get_driver(self):
        driver = get_driver(drivers=DRIVERS, provider=Provider.DUMMY)
        self.assertTrue(driver is not None)

        try:
            driver = get_driver(drivers=DRIVERS, provider="fooba")
        except AttributeError:
            pass
        else:
            self.fail("Invalid provider, but an exception was not thrown")

    def test_get_driver_string_and_enum_notation(self):
        driver = get_driver(drivers=DRIVERS, provider=Provider.DUMMY)
        self.assertEqual(driver, DummyNodeDriver)

        driver = get_driver(drivers=DRIVERS, provider="dummy")
        self.assertEqual(driver, DummyNodeDriver)

        driver = get_driver(drivers=DRIVERS, provider="DUMMY")
        self.assertEqual(driver, DummyNodeDriver)

    def test_set_driver(self):
        # Set an existing driver
        try:
            driver = set_driver(
                DRIVERS,
                Provider.DUMMY,
                "libcloud.storage.drivers.dummy",
                "DummyStorageDriver",
            )
        except AttributeError:
            pass

        # Register a new driver
        driver = set_driver(
            DRIVERS,
            "testingset",
            "libcloud.storage.drivers.dummy",
            "DummyStorageDriver",
        )

        self.assertTrue(driver is not None)

        # Register it again
        try:
            set_driver(
                DRIVERS,
                "testingset",
                "libcloud.storage.drivers.dummy",
                "DummyStorageDriver",
            )
        except AttributeError:
            pass

        # Register an invalid module
        try:
            set_driver(
                DRIVERS,
                "testingnew",
                "libcloud.storage.drivers.dummy1",
                "DummyStorageDriver",
            )
        except ImportError:
            pass

        # Register an invalid class
        try:
            set_driver(
                DRIVERS,
                "testingnew",
                "libcloud.storage.drivers.dummy",
                "DummyStorageDriver1",
            )
        except AttributeError:
            pass

    def test_deprecated_warning(self):
        warnings.showwarning = show_warning

        libcloud.utils.SHOW_DEPRECATION_WARNING = False
        self.assertEqual(len(WARNINGS_BUFFER), 0)
        libcloud.utils.deprecated_warning("test_module")
        self.assertEqual(len(WARNINGS_BUFFER), 0)

        libcloud.utils.SHOW_DEPRECATION_WARNING = True
        self.assertEqual(len(WARNINGS_BUFFER), 0)
        libcloud.utils.deprecated_warning("test_module")
        self.assertEqual(len(WARNINGS_BUFFER), 1)

    def test_in_development_warning(self):
        warnings.showwarning = show_warning

        libcloud.utils.SHOW_IN_DEVELOPMENT_WARNING = False
        self.assertEqual(len(WARNINGS_BUFFER), 0)
        libcloud.utils.in_development_warning("test_module")
        self.assertEqual(len(WARNINGS_BUFFER), 0)

        libcloud.utils.SHOW_IN_DEVELOPMENT_WARNING = True
        self.assertEqual(len(WARNINGS_BUFFER), 0)
        libcloud.utils.in_development_warning("test_module")
        self.assertEqual(len(WARNINGS_BUFFER), 1)

    def test_read_in_chunks_iterator_no_data(self):
        iterator = DummyIterator()
        generator1 = libcloud.utils.files.read_in_chunks(iterator=iterator, yield_empty=False)
        generator2 = libcloud.utils.files.read_in_chunks(iterator=iterator, yield_empty=True)

        # yield_empty=False
        count = 0
        for data in generator1:
            count += 1
            self.assertEqual(data, b(""))

        self.assertEqual(count, 0)

        # yield_empty=True
        count = 0
        for data in generator2:
            count += 1
            self.assertEqual(data, b(""))

        self.assertEqual(count, 1)

    def test_read_in_chunks_iterator(self):
        def iterator():
            for x in range(0, 1000):
                yield "aa"

        chunk_count = 0
        for result in libcloud.utils.files.read_in_chunks(
            iterator(), chunk_size=10, fill_size=False
        ):
            chunk_count += 1
            self.assertEqual(result, b("aa"))
        self.assertEqual(chunk_count, 1000)

        chunk_count = 0
        for result in libcloud.utils.files.read_in_chunks(
            iterator(), chunk_size=10, fill_size=True
        ):
            chunk_count += 1
            self.assertEqual(result, b("a") * 10)
        self.assertEqual(chunk_count, 200)

    def test_read_in_chunks_large_iterator_batches(self):
        def iterator():
            for x in range(0, 10):
                yield "a" * 10_000

        chunk_count = 0
        for result in libcloud.utils.files.read_in_chunks(
            iterator(), chunk_size=10, fill_size=False
        ):
            chunk_count += 1
            self.assertEqual(result, b("a") * 10_000)
        self.assertEqual(chunk_count, 10)

        chunk_count = 0
        for result in libcloud.utils.files.read_in_chunks(
            iterator(), chunk_size=10, fill_size=True
        ):
            chunk_count += 1
            self.assertEqual(result, b("a") * 10)
        self.assertEqual(chunk_count, 10_000)

    def test_read_in_chunks_filelike(self):
        class FakeFile(file):
            def __init__(self):
                self.remaining = 500

            def read(self, size):
                self.remaining -= 1
                if self.remaining == 0:
                    return ""
                return "b" * (size + 1)

        for index, result in enumerate(
            libcloud.utils.files.read_in_chunks(FakeFile(), chunk_size=10, fill_size=False)
        ):
            self.assertEqual(result, b("b" * 11))

        self.assertEqual(index, 498)

        for index, result in enumerate(
            libcloud.utils.files.read_in_chunks(FakeFile(), chunk_size=10, fill_size=True)
        ):
            if index != 548:
                self.assertEqual(result, b("b" * 10))
            else:
                self.assertEqual(result, b("b" * 9))

        self.assertEqual(index, 548)

    def test_read_in_chunks_corectness(self):
        data = "".join(
            random.choice(string.ascii_lowercase) for i in range(5 * 1024 * 1024)
        ).encode("utf-8")

        # fill_size = False
        chunk_size = None
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=False
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=False
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 2 * 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=False
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 5 * 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=False
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 10 * 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=False
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        # fill_size = True
        chunk_size = None
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=True
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=True
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 2 * 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=True
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 5 * 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=True
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

        chunk_size = 10 * 1024 * 1024
        iterator = libcloud.utils.files.read_in_chunks(
            iter([data]), chunk_size=chunk_size, fill_size=True
        )
        self.assertEqual(data, libcloud.utils.files.exhaust_iterator(iterator))

    def test_exhaust_iterator(self):
        def iterator_func():
            for x in range(0, 1000):
                yield "aa"

        data = b("aa" * 1000)
        iterator = libcloud.utils.files.read_in_chunks(iterator=iterator_func())
        result = libcloud.utils.files.exhaust_iterator(iterator=iterator)
        self.assertEqual(result, data)

        result = libcloud.utils.files.exhaust_iterator(iterator=iterator_func())
        self.assertEqual(result, data)

        data = "12345678990"
        iterator = StringIO(data)
        result = libcloud.utils.files.exhaust_iterator(iterator=iterator)
        self.assertEqual(result, b(data))

    def test_exhaust_iterator_empty_iterator(self):
        data = ""
        iterator = StringIO(data)
        result = libcloud.utils.files.exhaust_iterator(iterator=iterator)
        self.assertEqual(result, b(data))

    def test_unicode_urlquote(self):
        # Regression tests for LIBCLOUD-429
        # Note: this is a unicode literal
        val = "\xe9"

        uri = urlquote(val)
        self.assertEqual(b(uri), b("%C3%A9"))

        # Unicode without unicode characters
        uri = urlquote("v=1")
        self.assertEqual(b(uri), b("v%3D1"))

        # Already-encoded bytestring without unicode characters
        uri = urlquote(b("v=1"))
        self.assertEqual(b(uri), b("v%3D1"))

    def test_get_secure_random_string(self):
        for i in range(1, 500):
            value = get_secure_random_string(size=i)
            self.assertEqual(len(value), i)

    def test_hexadigits(self):
        self.assertEqual(hexadigits(b("")), [])
        self.assertEqual(hexadigits(b("a")), ["61"])
        self.assertEqual(hexadigits(b("AZaz09-")), ["41", "5a", "61", "7a", "30", "39", "2d"])

    def test_bchr(self):
        self.assertEqual(bchr(0), b"\x00")
        self.assertEqual(bchr(97), b"a")


class NetworkingUtilsTestCase(unittest.TestCase):
    def test_is_public_and_is_private_subnet(self):
        public_ips = ["213.151.0.8", "86.87.86.1", "8.8.8.8", "8.8.4.4"]

        private_ips = ["192.168.1.100", "10.0.0.1", "172.16.0.0"]

        for address in public_ips:
            is_public = is_public_subnet(ip=address)
            is_private = is_private_subnet(ip=address)

            self.assertTrue(is_public)
            self.assertFalse(is_private)

        for address in private_ips:
            is_public = is_public_subnet(ip=address)
            is_private = is_private_subnet(ip=address)

            self.assertFalse(is_public)
            self.assertTrue(is_private)

    def test_is_valid_ip_address(self):
        valid_ipv4_addresses = [
            "192.168.1.100",
            "10.0.0.1",
            "213.151.0.8",
            "77.77.77.77",
        ]

        invalid_ipv4_addresses = [
            "10.1",
            "256.256.256.256",
            "0.567.567.567",
            "192.168.0.257",
        ]

        valid_ipv6_addresses = [
            "fe80::200:5aee:feaa:20a2",
            "2607:f0d0:1002:51::4",
            "2607:f0d0:1002:0051:0000:0000:0000:0004",
            "::1",
        ]

        invalid_ipv6_addresses = [
            "2607:f0d",
            "2607:f0d0:0004",
        ]

        for address in valid_ipv4_addresses:
            status = is_valid_ip_address(address=address, family=socket.AF_INET)
            self.assertTrue(status)

        for address in valid_ipv6_addresses:
            status = is_valid_ip_address(address=address, family=socket.AF_INET6)
            self.assertTrue(status)

        for address in chain(invalid_ipv4_addresses, invalid_ipv6_addresses):
            status = is_valid_ip_address(address=address, family=socket.AF_INET)
            self.assertFalse(status)

        for address in chain(invalid_ipv4_addresses, invalid_ipv6_addresses):
            status = is_valid_ip_address(address=address, family=socket.AF_INET6)
            self.assertFalse(status)

    def test_join_ipv4_segments(self):
        values = [
            (("127", "0", "0", "1"), "127.0.0.1"),
            (("255", "255", "255", "0"), "255.255.255.0"),
        ]

        for segments, joined_ip in values:
            result = join_ipv4_segments(segments=segments)
            self.assertEqual(result, joined_ip)

    def test_increment_ipv4_segments(self):
        values = [
            (("127", "0", "0", "1"), "127.0.0.2"),
            (("255", "255", "255", "0"), "255.255.255.1"),
            (("254", "255", "255", "255"), "255.0.0.0"),
            (("100", "1", "0", "255"), "100.1.1.0"),
        ]

        for segments, incremented_ip in values:
            result = increment_ipv4_segments(segments=segments)
            result = join_ipv4_segments(segments=result)
            self.assertEqual(result, incremented_ip)


class TestPublicKeyUtils(unittest.TestCase):
    PUBKEY = (
        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDOfbWSXOlqvYjZmRO84/lIoV4gvuX+"
        "P1lLg50MMg6jZjLZIlYY081XPRmuom0xY0+BO++J2KgLl7gxJ6xMsKK2VQ+TakdfAH20"
        "XfMcTohd/zVCeWsbqZQvEhVXBo4hPIktcfNz0u9Ez3EtInO+kb7raLcRhOVi9QmOkOrC"
        "WtQU9mS71AWJuqI9H0YAnTiI8Hs5bn2tpMIqmTXT3g2bwywC25x1Nx9Hy0/FP+KUL6Ag"
        "vDXv47l+TgSDfTBEkvq+IF1ITrnaOG+nRE02oZC6cwHYTifM/IOollkujxIQmi2Z+j66"
        "OHSrjnEQugr0FqGJF2ygKfIh/i2u3fVLM60qE2NN user@example"
    )

    def test_pubkey_openssh_fingerprint(self):
        fp = get_pubkey_openssh_fingerprint(self.PUBKEY)
        self.assertEqual(fp, "35:22:13:5b:82:e2:5d:e1:90:8c:73:74:9f:ef:3b:d8")

    def test_pubkey_ssh2_fingerprint(self):
        fp = get_pubkey_ssh2_fingerprint(self.PUBKEY)
        self.assertEqual(fp, "11:ad:5d:4c:5b:99:c9:80:7e:81:03:76:5a:25:9d:8c")


def test_decorator():
    @wrap_non_libcloud_exceptions
    def foo():
        raise Exception("bork")

    with pytest.raises(LibcloudError):
        foo()


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.types import RecordType
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_DNSIMPLE
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.dns.drivers.dnsimple import DNSimpleDNSDriver


class DNSimpleDNSTests(unittest.TestCase):
    def setUp(self):
        DNSimpleDNSDriver.connectionCls.conn_class = DNSimpleDNSMockHttp
        DNSimpleDNSMockHttp.type = None
        self.driver = DNSimpleDNSDriver(*DNS_PARAMS_DNSIMPLE)

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary, 'key "%s" not in dictionary' % (key))

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 15)
        self.assertTrue(RecordType.A in record_types)
        self.assertTrue(RecordType.AAAA in record_types)
        self.assertTrue(RecordType.ALIAS in record_types)
        self.assertTrue(RecordType.CNAME in record_types)
        self.assertTrue(RecordType.HINFO in record_types)
        self.assertTrue(RecordType.MX in record_types)
        self.assertTrue(RecordType.NAPTR in record_types)
        self.assertTrue(RecordType.NS in record_types)
        self.assertTrue("POOL" in record_types)
        self.assertTrue(RecordType.SPF in record_types)
        self.assertTrue(RecordType.SOA in record_types)
        self.assertTrue(RecordType.SRV in record_types)
        self.assertTrue(RecordType.SSHFP in record_types)
        self.assertTrue(RecordType.TXT in record_types)
        self.assertTrue(RecordType.URL in record_types)

    def test_list_zones_success(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)

        zone1 = zones[0]
        self.assertEqual(zone1.id, "1")
        self.assertEqual(zone1.type, "master")
        self.assertEqual(zone1.domain, "example.com")
        self.assertEqual(zone1.ttl, 3600)
        self.assertHasKeys(
            zone1.extra,
            [
                "registrant_id",
                "user_id",
                "unicode_name",
                "token",
                "state",
                "language",
                "lockable",
                "auto_renew",
                "whois_protected",
                "record_count",
                "service_count",
                "expires_on",
                "created_at",
                "updated_at",
            ],
        )

        zone2 = zones[1]
        self.assertEqual(zone2.id, "2")
        self.assertEqual(zone2.type, "master")
        self.assertEqual(zone2.domain, "example.com")
        self.assertEqual(zone2.ttl, 3600)
        self.assertHasKeys(
            zone2.extra,
            [
                "registrant_id",
                "user_id",
                "unicode_name",
                "token",
                "state",
                "language",
                "lockable",
                "auto_renew",
                "whois_protected",
                "record_count",
                "service_count",
                "expires_on",
                "created_at",
                "updated_at",
            ],
        )

    def test_list_records_success(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 3)

        record1 = records[0]
        self.assertEqual(record1.id, "31")
        self.assertEqual(record1.name, "")
        self.assertEqual(record1.type, RecordType.A)
        self.assertEqual(record1.data, "1.2.3.4")
        self.assertHasKeys(
            record1.extra, ["ttl", "created_at", "updated_at", "domain_id", "priority"]
        )

        record2 = records[1]
        self.assertEqual(record2.id, "2")
        self.assertEqual(record2.name, "www")
        self.assertEqual(record2.type, RecordType.CNAME)
        self.assertEqual(record2.data, "example.com")
        self.assertHasKeys(
            record2.extra, ["ttl", "created_at", "updated_at", "domain_id", "priority"]
        )

        record3 = records[2]
        self.assertEqual(record3.id, "32")
        self.assertEqual(record3.name, "")
        self.assertEqual(record3.type, RecordType.MX)
        self.assertEqual(record3.data, "mail.example.com")
        self.assertHasKeys(
            record3.extra, ["ttl", "created_at", "updated_at", "domain_id", "priority"]
        )

    def test_get_zone_success(self):
        zone1 = self.driver.get_zone(zone_id="1")
        self.assertEqual(zone1.id, "1")
        self.assertEqual(zone1.type, "master")
        self.assertEqual(zone1.domain, "example.com")
        self.assertEqual(zone1.ttl, 3600)
        self.assertHasKeys(
            zone1.extra,
            [
                "registrant_id",
                "user_id",
                "unicode_name",
                "token",
                "state",
                "language",
                "lockable",
                "auto_renew",
                "whois_protected",
                "record_count",
                "service_count",
                "expires_on",
                "created_at",
                "updated_at",
            ],
        )

    def test_get_record_success(self):
        record = self.driver.get_record(zone_id="1", record_id="123")
        self.assertEqual(record.id, "123")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.type, RecordType.CNAME)
        self.assertEqual(record.data, "example.com")
        self.assertHasKeys(
            record.extra, ["ttl", "created_at", "updated_at", "domain_id", "priority"]
        )

    def test_create_zone_success(self):
        DNSimpleDNSMockHttp.type = "CREATE"
        zone = self.driver.create_zone(domain="example.com")
        self.assertEqual(zone.id, "1")
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.ttl, 3600)
        self.assertEqual(zone.type, "master")
        self.assertHasKeys(
            zone.extra,
            [
                "registrant_id",
                "user_id",
                "unicode_name",
                "token",
                "state",
                "language",
                "lockable",
                "auto_renew",
                "whois_protected",
                "record_count",
                "service_count",
                "expires_on",
                "created_at",
                "updated_at",
            ],
        )

    def test_create_record_success(self):
        zone = self.driver.list_zones()[0]
        DNSimpleDNSMockHttp.type = "CREATE"
        record = self.driver.create_record(
            name="domain4", zone=zone, type=RecordType.MX, data="mail.example.com"
        )
        self.assertEqual(record.id, "172")
        self.assertEqual(record.name, "")
        self.assertEqual(record.type, RecordType.MX)
        self.assertEqual(record.data, "mail.example.com")
        self.assertHasKeys(
            record.extra, ["ttl", "created_at", "updated_at", "domain_id", "priority"]
        )

    def test_update_record_success(self):
        record = self.driver.get_record(zone_id="1", record_id="123")
        DNSimpleDNSMockHttp.type = "UPDATE"
        extra = {"ttl": 4500}
        record1 = self.driver.update_record(
            record=record, name="www", type=record.type, data="updated.com", extra=extra
        )
        self.assertEqual(record.data, "example.com")
        self.assertEqual(record.extra.get("ttl"), 3600)
        self.assertEqual(record1.data, "updated.com")
        self.assertEqual(record1.extra.get("ttl"), 4500)

    def test_delete_zone_success(self):
        zone = self.driver.list_zones()[0]
        DNSimpleDNSMockHttp.type = "DELETE_200"
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_success_future_implementation(self):
        zone = self.driver.list_zones()[0]
        DNSimpleDNSMockHttp.type = "DELETE_204"
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_record_success(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 3)
        record = records[1]
        DNSimpleDNSMockHttp.type = "DELETE_200"
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_success_future_implementation(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 3)
        record = records[1]
        DNSimpleDNSMockHttp.type = "DELETE_204"
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)


class DNSimpleDNSMockHttp(MockHttp):
    fixtures = DNSFileFixtures("dnsimple")

    def _v1_domains(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_CREATE(self, method, url, body, headers):
        body = self.fixtures.load("create_domain.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1(self, method, url, body, headers):
        body = self.fixtures.load("get_domain.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records(self, method, url, body, headers):
        body = self.fixtures.load("list_domain_records.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_123(self, method, url, body, headers):
        body = self.fixtures.load("get_domain_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_CREATE(self, method, url, body, headers):
        body = self.fixtures.load("create_domain_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_123_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load("update_domain_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_domains_1_DELETE_200(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _v1_domains_1_DELETE_204(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.NO_CONTENT])

    def _v1_domains_1_records_2_DELETE_200(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _v1_domains_1_records_2_DELETE_204(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.NO_CONTENT])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.dns.types import RecordType
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_ONAPP
from libcloud.common.exceptions import BaseHTTPError
from libcloud.dns.drivers.onapp import OnAppDNSDriver
from libcloud.test.file_fixtures import DNSFileFixtures


class OnAppDNSTests(LibcloudTestCase):
    def setUp(self):
        OnAppDNSDriver.connectionCls.conn_class = OnAppDNSMockHttp
        OnAppDNSMockHttp.type = None
        self.driver = OnAppDNSDriver(*DNS_PARAMS_ONAPP)

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary, 'key "%s" not in dictionary' % (key))

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 8)
        self.assertTrue(RecordType.A in record_types)
        self.assertTrue(RecordType.AAAA in record_types)
        self.assertTrue(RecordType.CNAME in record_types)
        self.assertTrue(RecordType.MX in record_types)
        self.assertTrue(RecordType.NS in record_types)
        self.assertTrue(RecordType.SOA in record_types)
        self.assertTrue(RecordType.SRV in record_types)
        self.assertTrue(RecordType.TXT in record_types)

    def test_list_zones_success(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)

        zone1 = zones[0]
        self.assertEqual(zone1.id, "1")
        self.assertEqual(zone1.type, "master")
        self.assertEqual(zone1.domain, "example.com")
        self.assertEqual(zone1.ttl, 1200)
        self.assertHasKeys(zone1.extra, ["user_id", "cdn_reference", "created_at", "updated_at"])

        zone2 = zones[1]
        self.assertEqual(zone2.id, "2")
        self.assertEqual(zone2.type, "master")
        self.assertEqual(zone2.domain, "example.net")
        self.assertEqual(zone2.ttl, 1200)
        self.assertHasKeys(zone2.extra, ["user_id", "cdn_reference", "created_at", "updated_at"])

    def test_get_zone_success(self):
        zone1 = self.driver.get_zone(zone_id="1")
        self.assertEqual(zone1.id, "1")
        self.assertEqual(zone1.type, "master")
        self.assertEqual(zone1.domain, "example.com")
        self.assertEqual(zone1.ttl, 1200)
        self.assertHasKeys(zone1.extra, ["user_id", "cdn_reference", "created_at", "updated_at"])

    def test_get_zone_not_found(self):
        OnAppDNSMockHttp.type = "NOT_FOUND"
        try:
            self.driver.get_zone(zone_id="3")
        except BaseHTTPError:
            self.assertRaises(Exception)

    def test_create_zone_success(self):
        OnAppDNSMockHttp.type = "CREATE"
        zone = self.driver.create_zone(domain="example.com")
        self.assertEqual(zone.id, "1")
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.ttl, 1200)
        self.assertEqual(zone.type, "master")
        self.assertHasKeys(zone.extra, ["user_id", "cdn_reference", "created_at", "updated_at"])

    def test_delete_zone(self):
        zone = self.driver.get_zone(zone_id="1")
        OnAppDNSMockHttp.type = "DELETE"
        self.assertTrue(self.driver.delete_zone(zone))

    def test_list_records_success(self):
        zone = self.driver.get_zone(zone_id="1")
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 5)

        record1 = records[0]
        self.assertEqual(record1.id, "111222")
        self.assertEqual(record1.name, "@")
        self.assertEqual(record1.type, RecordType.A)
        self.assertEqual(record1.ttl, 3600)
        self.assertEqual(record1.data["ip"], "123.156.189.1")

        record2 = records[2]
        self.assertEqual(record2.id, "111224")
        self.assertEqual(record2.name, "mail")
        self.assertEqual(record1.ttl, 3600)
        self.assertEqual(record2.type, RecordType.CNAME)
        self.assertEqual(record2.data["hostname"], "examplemail.com")

        record3 = records[4]
        self.assertEqual(record3.id, "111226")
        self.assertEqual(record3.name, "@")
        self.assertEqual(record3.type, RecordType.MX)
        self.assertEqual(record3.data["hostname"], "mx2.examplemail.com")

    def test_get_record_success(self):
        record = self.driver.get_record(zone_id="1", record_id="123")
        self.assertEqual(record.id, "123")
        self.assertEqual(record.name, "@")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data["ip"], "123.156.189.1")

    def test_create_record_success(self):
        zone = self.driver.get_zone(zone_id="1")
        OnAppDNSMockHttp.type = "CREATE"
        record = self.driver.create_record(
            name="blog", zone=zone, type=RecordType.A, data="123.156.189.2"
        )
        self.assertEqual(record.id, "111227")
        self.assertEqual(record.name, "blog")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data["ip"], "123.156.189.2")
        self.assertEqual(record.data["ttl"], 3600)

    def test_update_record_success(self):
        record = self.driver.get_record(zone_id="1", record_id="123")
        OnAppDNSMockHttp.type = "UPDATE"
        extra = {"ttl": 4500}
        record1 = self.driver.update_record(
            record=record, name="@", type=record.type, data="123.156.189.2", extra=extra
        )
        self.assertEqual(record.data["ip"], "123.156.189.1")
        self.assertEqual(record.ttl, 3600)
        self.assertEqual(record1.data["ip"], "123.156.189.2")
        self.assertEqual(record1.ttl, 4500)

    def test_delete_record_success(self):
        record = self.driver.get_record(zone_id="1", record_id="123")
        OnAppDNSMockHttp.type = "DELETE"
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)


class OnAppDNSMockHttp(MockHttp):
    fixtures = DNSFileFixtures("onapp")

    def _dns_zones_json(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _dns_zones_1_json(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _dns_zones_3_json_NOT_FOUND(self, method, url, body, headers):
        body = self.fixtures.load("dns_zone_not_found.json")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _dns_zones_json_CREATE(self, method, url, body, headers):
        body = self.fixtures.load("create_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _dns_zones_1_json_DELETE(self, method, url, body, headers):
        return (httplib.NO_CONTENT, "", {}, httplib.responses[httplib.NO_CONTENT])

    def _dns_zones_1_records_json(self, method, url, body, headers):
        body = self.fixtures.load("list_records.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _dns_zones_1_records_123_json(self, method, url, body, headers):
        body = self.fixtures.load("get_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _dns_zones_1_records_json_CREATE(self, method, url, body, headers):
        body = self.fixtures.load("create_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _dns_zones_1_records_123_json_UPDATE(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("get_record_after_update.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:
            return (httplib.NO_CONTENT, "", {}, httplib.responses[httplib.NO_CONTENT])

    def _dns_zones_1_json_UPDATE(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _dns_zones_1_records_123_json_DELETE(self, method, url, body, headers):
        return (httplib.NO_CONTENT, "", {}, httplib.responses[httplib.NO_CONTENT])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.types import LibcloudError
from libcloud.storage.providers import Provider
from libcloud.storage.drivers.cloudfiles import CloudFilesConnection, CloudFilesStorageDriver

KTUCLOUDSTORAGE_AUTH_URL = "https://ssproxy.ucloudbiz.olleh.com/auth/v1.0"
KTUCLOUDSTORAGE_API_VERSION = "1.0"


class KTUCloudStorageConnection(CloudFilesConnection):
    """
    Connection class for the KT UCloud Storage endpoint.
    """

    auth_url = KTUCLOUDSTORAGE_AUTH_URL
    _auth_version = KTUCLOUDSTORAGE_API_VERSION

    def get_endpoint(self):
        eps = self.service_catalog.get_endpoints(name="cloudFiles")

        if len(eps) == 0:
            raise LibcloudError("Could not find specified endpoint")

        ep = eps[0]
        public_url = ep.url

        if not public_url:
            raise LibcloudError("Could not find specified endpoint")

        return public_url


class KTUCloudStorageDriver(CloudFilesStorageDriver):
    """
    Cloudfiles storage driver for the UK endpoint.
    """

    type = Provider.KTUCLOUD
    name = "KTUCloud Storage"
    connectionCls = KTUCloudStorageConnection



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest
from unittest.mock import MagicMock

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.dns.base import Zone, Record
from libcloud.dns.types import (
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
)
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_DURABLEDNS
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.dns.drivers.durabledns import (
    DEFAULT_TTL,
    ZONE_EXTRA_PARAMS_DEFAULT_VALUES,
    RECORD_EXTRA_PARAMS_DEFAULT_VALUES,
    DurableDNSDriver,
)


class DurableDNSTests(LibcloudTestCase):
    def setUp(self):
        DurableDNSDriver.connectionCls.conn_class = DurableDNSMockHttp
        DurableDNSMockHttp.type = None
        self.driver = DurableDNSDriver(*DNS_PARAMS_DURABLEDNS)

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary, 'key "%s" not in dictionary' % (key))

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 10)
        self.assertTrue(RecordType.A in record_types)
        self.assertTrue(RecordType.AAAA in record_types)
        self.assertTrue(RecordType.CNAME in record_types)
        self.assertTrue(RecordType.HINFO in record_types)
        self.assertTrue(RecordType.MX in record_types)
        self.assertTrue(RecordType.NS in record_types)
        self.assertTrue(RecordType.PTR in record_types)
        self.assertTrue(RecordType.RP in record_types)
        self.assertTrue(RecordType.SRV in record_types)
        self.assertTrue(RecordType.TXT in record_types)

    def test_list_zones(self):
        extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "serial": "1437473456",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=extra,
        )
        self.driver.get_zone = MagicMock(return_value=zone)
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)
        zone = zones[0]
        self.assertEqual(zone.id, "myzone.com.")
        self.assertEqual(zone.domain, "myzone.com.")
        self.assertEqual(zone.ttl, 1300)
        self.assertEqual(zone.extra["ns"], "ns1.durabledns.com.")
        self.assertEqual(zone.extra["mbox"], "mail.myzone.com")
        self.assertEqual(zone.extra["serial"], "1437473456")
        self.assertEqual(zone.extra["refresh"], "13000")
        self.assertEqual(zone.extra["retry"], 7200)
        self.assertEqual(zone.extra["expire"], 1300)
        self.assertEqual(zone.extra["minimum"], 13)
        self.assertEqual(zone.extra["xfer"], "127.0.0.1")
        self.assertEqual(zone.extra["update_acl"], "127.0.0.1")
        self.assertEqual(len(zone.extra.keys()), 9)

    def test_list_records(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        extra = {"aux": 1, "ttl": 3600}
        record = Record(
            id="353286987",
            type="A",
            zone=zone,
            name="record1",
            data="192.168.0.1",
            driver=self,
            extra=extra,
        )
        self.driver.get_record = MagicMock(return_value=record)
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 2)
        self.assertEqual(record.id, "353286987")
        self.assertEqual(record.name, "record1")
        self.assertEqual(record.type, "A")
        self.assertEqual(record.data, "192.168.0.1")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.extra["aux"], 1)
        self.assertEqual(record.extra["ttl"], 3600)

    def test_list_records_zone_does_not_exist(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.list_records(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_get_zone(self):
        zone = self.driver.get_zone(zone_id="myzone.com.")
        self.assertEqual(zone.id, "myzone.com.")
        self.assertEqual(zone.domain, "myzone.com.")
        self.assertEqual(zone.ttl, 1300)
        self.assertEqual(zone.extra["ns"], "ns1.durabledns.com.")
        self.assertEqual(zone.extra["mbox"], "mail.myzone.com")
        self.assertEqual(zone.extra["serial"], "1437473456")
        self.assertEqual(zone.extra["refresh"], "13000")
        self.assertEqual(zone.extra["retry"], 7200)
        self.assertEqual(zone.extra["expire"], 1300)
        self.assertEqual(zone.extra["minimum"], 13)
        self.assertEqual(zone.extra["xfer"], "127.0.0.1/32")
        self.assertEqual(zone.extra["update_acl"], "127.0.0.1/32,127.0.0.100/32")
        self.assertEqual(len(zone.extra.keys()), 9)

    def test_get_zone_does_not_exist(self):
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.get_zone(zone_id="nonexistentzone.com.")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "nonexistentzone.com.")
        else:
            self.fail("Exception was not thrown")

    def test_get_record(self):
        record = self.driver.get_record(zone_id="myzone.com.", record_id="record1")
        self.assertEqual(record.id, "353286987")
        self.assertEqual(record.name, "record1")
        self.assertEqual(record.type, "A")
        self.assertEqual(record.data, "192.168.0.1")
        self.assertEqual(record.extra["aux"], 1)
        self.assertEqual(record.extra["ttl"], 3600)

    def test_get_record_zone_does_not_exist(self):
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.get_record(zone_id="nonexistentzone.com.", record_id="record1")
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_record_record_does_not_exist(self):
        DurableDNSMockHttp.type = "RECORD_DOES_NOT_EXIST"
        try:
            self.driver.get_record(zone_id="", record_id="")
        except RecordDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_zone_with_extra_param(self):
        DurableDNSMockHttp.type = "WITH_EXTRA_PARAMS"
        zone = self.driver.create_zone(
            domain="myzone.com.",
            ttl=4000,
            extra={"mbox": "mail.myzone.com", "minimum": 50000},
        )
        extra = ZONE_EXTRA_PARAMS_DEFAULT_VALUES
        self.assertEqual(zone.id, "myzone.com.")
        self.assertEqual(zone.domain, "myzone.com.")
        self.assertEqual(zone.ttl, 4000)
        self.assertEqual(zone.extra["ns"], extra["ns"])
        self.assertEqual(zone.extra["mbox"], "mail.myzone.com")
        self.assertEqual(zone.extra["serial"], "1437473456")
        self.assertEqual(zone.extra["refresh"], extra["refresh"])
        self.assertEqual(zone.extra["retry"], extra["retry"])
        self.assertEqual(zone.extra["expire"], extra["expire"])
        self.assertEqual(zone.extra["minimum"], 50000)
        self.assertEqual(zone.extra["xfer"], extra["xfer"])
        self.assertEqual(zone.extra["update_acl"], extra["update_acl"])
        self.assertEqual(len(zone.extra.keys()), 9)

    def test_create_zone_no_extra_param(self):
        DurableDNSMockHttp.type = "NO_EXTRA_PARAMS"
        zone = self.driver.create_zone(domain="myzone.com.")
        extra = ZONE_EXTRA_PARAMS_DEFAULT_VALUES
        self.assertEqual(zone.id, "myzone.com.")
        self.assertEqual(zone.domain, "myzone.com.")
        self.assertEqual(zone.ttl, DEFAULT_TTL)
        self.assertEqual(zone.extra["ns"], extra["ns"])
        self.assertEqual(zone.extra["mbox"], extra["mbox"])
        self.assertEqual(zone.extra["serial"], "1437473456")
        self.assertEqual(zone.extra["refresh"], extra["refresh"])
        self.assertEqual(zone.extra["retry"], extra["retry"])
        self.assertEqual(zone.extra["expire"], extra["expire"])
        self.assertEqual(zone.extra["minimum"], extra["minimum"])
        self.assertEqual(zone.extra["xfer"], extra["xfer"])
        self.assertEqual(zone.extra["update_acl"], extra["update_acl"])
        self.assertEqual(len(zone.extra.keys()), 9)

    def test_create_zone_zone_already_exist(self):
        DurableDNSMockHttp.type = "ZONE_ALREADY_EXIST"
        try:
            self.driver.create_zone(domain="myzone.com.")
        except ZoneAlreadyExistsError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_record_no_extra_param(self):
        zone = self.driver.list_zones()[0]
        DurableDNSMockHttp.type = "NO_EXTRA_PARAMS"
        record = self.driver.create_record(
            name="record1", zone=zone, type=RecordType.A, data="1.2.3.4"
        )

        self.assertEqual(record.id, "353367855")
        self.assertEqual(record.name, "record1")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "1.2.3.4")
        self.assertEqual(record.extra.get("aux"), RECORD_EXTRA_PARAMS_DEFAULT_VALUES.get("aux"))
        self.assertEqual(record.extra.get("ttl"), RECORD_EXTRA_PARAMS_DEFAULT_VALUES.get("ttl"))

    def test_create_record_with_extra_param(self):
        zone = self.driver.list_zones()[0]
        DurableDNSMockHttp.type = "WITH_EXTRA_PARAMS"
        record = self.driver.create_record(
            name="record1",
            zone=zone,
            type=RecordType.A,
            data="1.2.3.4",
            extra={"ttl": 4000},
        )

        self.assertEqual(record.id, "353367855")
        self.assertEqual(record.name, "record1")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "1.2.3.4")
        self.assertEqual(record.extra.get("aux"), RECORD_EXTRA_PARAMS_DEFAULT_VALUES.get("aux"))
        self.assertEqual(record.extra.get("ttl"), 4000)

    def test_create_record_zone_does_not_exist(self):
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="deletedzone.com.",
            domain="deletedzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        try:
            self.driver.create_record(name="record1", zone=zone, type=RecordType.A, data="1.2.3.4")
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_update_zone(self):
        # We'll assume that this zone has been created before. So will have
        # a serial number in his extra attributes. Later we are going to
        # check that after the update, serial number should change to new one.
        DurableDNSMockHttp.type = "UPDATE_ZONE"
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "serial": "1437473456",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="deletedzone.com.",
            domain="deletedzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        new_extra = {"minimum": 5000, "expire": 8000}
        updated_zone = self.driver.update_zone(
            zone, zone.domain, type=zone.type, ttl=4000, extra=new_extra
        )

        self.assertEqual(updated_zone.id, "myzone.com.")
        self.assertEqual(updated_zone.domain, "myzone.com.")
        self.assertEqual(updated_zone.ttl, 4000)
        self.assertEqual(updated_zone.extra["ns"], z_extra["ns"])
        self.assertEqual(updated_zone.extra["mbox"], z_extra["mbox"])
        self.assertEqual(updated_zone.extra["serial"], "1437475078")
        self.assertEqual(updated_zone.extra["refresh"], z_extra["refresh"])
        self.assertEqual(updated_zone.extra["retry"], z_extra["retry"])
        self.assertEqual(updated_zone.extra["expire"], 8000)
        self.assertEqual(updated_zone.extra["minimum"], 5000)
        self.assertEqual(updated_zone.extra["xfer"], z_extra["xfer"])
        self.assertEqual(updated_zone.extra["update_acl"], z_extra["update_acl"])
        self.assertEqual(len(updated_zone.extra.keys()), 9)

    def test_update_zone_zone_does_not_exist(self):
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "serial": "1437473456",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="deletedzone.com.",
            domain="deletedzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        try:
            self.driver.update_zone(zone, zone.domain)
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_update_record(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        extra = {"aux": 1, "ttl": 3600}
        record = Record(
            id="353286987",
            type="A",
            zone=zone,
            name="record1",
            data="192.168.0.1",
            driver=self,
            extra=extra,
        )
        new_extra = {"aux": 0, "ttl": 4500}
        updated_record = self.driver.update_record(
            record, record.name, record.type, record.data, extra=new_extra
        )

        self.assertEqual(updated_record.data, "192.168.0.1")
        self.assertEqual(updated_record.id, "353286987")
        self.assertEqual(updated_record.name, "record1")
        self.assertEqual(updated_record.zone, record.zone)
        self.assertEqual(updated_record.type, RecordType.A)
        self.assertEqual(updated_record.extra.get("aux"), 0)
        self.assertEqual(updated_record.extra.get("ttl"), 4500)

    def test_update_record_zone_does_not_exist(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        extra = {"aux": 1, "ttl": 3600}
        record = Record(
            id="353286987",
            type="A",
            zone=zone,
            name="record1",
            data="192.168.0.1",
            driver=self,
            extra=extra,
        )
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.update_record(record, record.name, record.type, record.data)
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_zone(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_zone_does_not_exist(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_record(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        extra = {"aux": 1, "ttl": 3600}
        record = Record(
            id="353286987",
            type="A",
            zone=zone,
            name="record1",
            data="192.168.0.1",
            driver=self,
            extra=extra,
        )
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_record_does_not_exist(self):
        z_extra = {
            "ns": "ns1.durabledns.com.",
            "mbox": "mail.myzone.com",
            "refresh": "13000",
            "retry": 7200,
            "expire": 1300,
            "minimum": 13,
            "xfer": "127.0.0.1",
            "update_acl": "127.0.0.1",
        }
        zone = Zone(
            id="myzone.com.",
            domain="myzone.com.",
            type="master",
            ttl=1300,
            driver=self.driver,
            extra=z_extra,
        )
        extra = {"aux": 1, "ttl": 3600}
        record = Record(
            id="353286987",
            type="A",
            zone=zone,
            name="record1",
            data="192.168.0.1",
            driver=self,
            extra=extra,
        )
        DurableDNSMockHttp.type = "RECORD_DOES_NOT_EXIST"
        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_record_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        DurableDNSMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.delete_record(record=record)
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")


class DurableDNSMockHttp(MockHttp):
    fixtures = DNSFileFixtures("durabledns")

    def _services_dns_listZones_php(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_listRecords_php(self, method, url, body, headers):
        body = self.fixtures.load("list_records.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_listRecords_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_records_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getZone_php(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getZone_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getRecord_php(self, method, url, body, headers):
        body = self.fixtures.load("get_record.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getRecord_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_record_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getRecord_php_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_record_RECORD_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_createZone_php_WITH_EXTRA_PARAMS(self, method, url, body, headers):
        body = self.fixtures.load("create_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getZone_php_WITH_EXTRA_PARAMS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_WITH_EXTRA_PARAMS.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_createZone_php_NO_EXTRA_PARAMS(self, method, url, body, headers):
        body = self.fixtures.load("create_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getZone_php_NO_EXTRA_PARAMS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_NO_EXTRA_PARAMS.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_createZone_php_ZONE_ALREADY_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("create_zone_ZONE_ALREADY_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_createRecord_php_NO_EXTRA_PARAMS(self, method, url, body, headers):
        body = self.fixtures.load("create_record_NO_EXTRA_PARAMS.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_createRecord_php_WITH_EXTRA_PARAMS(self, method, url, body, headers):
        body = self.fixtures.load("create_record_WITH_EXTRA_PARAMS.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_createRecord_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("create_record_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_updateZone_php_UPDATE_ZONE(self, method, url, body, headers):
        body = self.fixtures.load("update_zone_UPDATE_ZONE.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_getZone_php_UPDATE_ZONE(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_UPDATE_ZONE.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_updateZone_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("update_zone_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_updateRecord_php(self, method, url, body, headers):
        body = self.fixtures.load("update_record.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_updateRecord_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("update_record_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_deleteZone_php(self, method, url, body, headers):
        body = self.fixtures.load("delete_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_deleteZone_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("delete_zone_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_deleteRecord_php(self, method, url, body, headers):
        body = self.fixtures.load("delete_record.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_deleteRecord_php_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("delete_record_RECORD_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _services_dns_deleteRecord_php_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("delete_record_ZONE_DOES_NOT_EXIST.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.base import Zone, Record
from libcloud.dns.types import (
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
)
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import VULTR_PARAMS
from libcloud.dns.drivers.vultr import VultrDNSDriver, VultrDNSDriverV1
from libcloud.test.file_fixtures import DNSFileFixtures


class VultrTests(unittest.TestCase):
    def setUp(self):
        VultrMockHttp.type = None
        VultrDNSDriverV1.connectionCls.conn_class = VultrMockHttp
        self.driver = VultrDNSDriver(*VULTR_PARAMS, api_version="1")
        self.test_zone = Zone(
            id="test.com",
            type="master",
            ttl=None,
            domain="test.com",
            extra={},
            driver=self,
        )
        self.test_record = Record(
            id="31",
            type=RecordType.A,
            name="test",
            zone=self.test_zone,
            data="127.0.0.1",
            driver=self,
            extra={},
        )

    def test_correct_class_is_used(self):
        self.assertIsInstance(self.driver, VultrDNSDriverV1)

    def test_list_zones_empty(self):
        VultrMockHttp.type = "EMPTY_ZONES_LIST"
        zones = self.driver.list_zones()

        self.assertEqual(zones, [])

    def test_list_zones_success(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 4)

        zone = zones[0]
        self.assertEqual(zone.id, "example.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "example.com")
        self.assertIsNone(zone.ttl)

        zone = zones[1]
        self.assertEqual(zone.id, "zupo.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "zupo.com")
        self.assertIsNone(zone.ttl)

        zone = zones[2]
        self.assertEqual(zone.id, "oltjano.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "oltjano.com")
        self.assertIsNone(zone.ttl)

        zone = zones[3]
        self.assertEqual(zone.id, "13.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "13.com")
        self.assertIsNone(zone.ttl)

    def test_get_zone_zone_does_not_exist(self):
        VultrMockHttp.type = "GET_ZONE_ZONE_DOES_NOT_EXIST"
        try:
            self.driver.get_zone(zone_id="test.com")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "test.com")
        else:
            self.fail("Exception was not thrown")

    def test_get_zone_success(self):
        VultrMockHttp.type = "GET_ZONE_SUCCESS"
        zone = self.driver.get_zone(zone_id="zupo.com")

        self.assertEqual(zone.id, "zupo.com")
        self.assertEqual(zone.domain, "zupo.com")
        self.assertEqual(zone.type, "master")
        self.assertIsNone(zone.ttl)

    def test_delete_zone_zone_does_not_exist(self):
        VultrMockHttp.type = "DELETE_ZONE_ZONE_DOES_NOT_EXIST"

        try:
            self.driver.delete_zone(zone=self.test_zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, self.test_zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_zone_success(self):
        zone = self.driver.list_zones()[0]
        status = self.driver.delete_zone(zone=zone)

        self.assertTrue(status)

    def test_create_zone_success(self):
        zone = self.driver.create_zone(domain="test.com", extra={"serverip": "127.0.0.1"})

        self.assertEqual(zone.id, "test.com")
        self.assertEqual(zone.domain, "test.com")
        self.assertEqual(zone.type, "master"),
        self.assertIsNone(zone.ttl)

    def test_create_zone_zone_already_exists(self):
        VultrMockHttp.type = "CREATE_ZONE_ZONE_ALREADY_EXISTS"

        try:
            self.driver.create_zone(domain="example.com", extra={"serverip": "127.0.0.1"})
        except ZoneAlreadyExistsError as e:
            self.assertEqual(e.zone_id, "example.com")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_record_does_not_exist(self):
        VultrMockHttp.type = "GET_RECORD_RECORD_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="zupo.com", record_id="1300")
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, "1300")
        else:
            self.fail("Exception was not thrown")

    def test_list_records_zone_does_not_exist(self):
        VultrMockHttp.type = "LIST_RECORDS_ZONE_DOES_NOT_EXIST"

        try:
            self.driver.list_records(zone=self.test_zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, self.test_zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_list_records_empty(self):
        VultrMockHttp.type = "EMPTY_RECORDS_LIST"
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)

        self.assertEqual(records, [])

    def test_list_records_success(self):
        zone = self.driver.get_zone(zone_id="zupo.com")
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 2)

        arecord = records[0]
        self.assertEqual(arecord.id, "13")
        self.assertEqual(arecord.name, "arecord")
        self.assertEqual(arecord.type, RecordType.A)
        self.assertEqual(arecord.data, "127.0.0.1")

    def test_get_record_success(self):
        VultrMockHttp.type = "GET_RECORD"
        record = self.driver.get_record(zone_id="zupo.com", record_id="1300")

        self.assertEqual(record.id, "1300")
        self.assertEqual(record.name, "zupo")
        self.assertEqual(record.data, "127.0.0.1")
        self.assertEqual(record.type, RecordType.A)

    def test_delete_record_record_does_not_exist(self):
        VultrMockHttp.type = "DELETE_RECORD_RECORD_DOES_NOT_EXIST"

        try:
            self.driver.delete_record(record=self.test_record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, self.test_record.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        status = self.driver.delete_record(record=record)

        self.assertTrue(status)


class VultrMockHttp(MockHttp):
    fixtures = DNSFileFixtures("vultr")

    def _v1_dns_list(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_records(self, method, url, body, headers):
        body = self.fixtures.load("list_records.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_EMPTY_ZONES_LIST(self, method, url, body, headers):
        body = self.fixtures.load("empty_zones_list.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_GET_ZONE_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_GET_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_EMPTY_RECORDS_LIST(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_records_EMPTY_RECORDS_LIST(self, method, url, body, headers):
        body = self.fixtures.load("empty_records_list.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_GET_RECORD(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_records_GET_RECORD(self, method, url, body, headers):
        body = self.fixtures.load("get_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_GET_RECORD_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_records_GET_RECORD_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_records.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_delete_domain(self, method, url, body, headers):
        body = ""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_delete_record(self, method, url, body, headers):
        body = ""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_create_domain(self, method, url, body, headers):
        body = ""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_CREATE_ZONE_ZONE_ALREADY_EXISTS(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_create_domain_CREATE_ZONE_ZONE_ALREADY_EXISTS(self, method, url, body, headers):
        body = ""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_DELETE_ZONE_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_delete_domain_DELETE_ZONE_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_records_DELETE_RECORD_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_records.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_delete_record_DELETE_RECORD_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_DELETE_RECORD_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("test_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_list_LIST_RECORDS_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("list_domains.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_dns_records_LIST_RECORDS_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import unittest
from unittest import mock

from libcloud.test import MockHttp  # pylint: disable-msg=E0611
from libcloud.test import make_response, generate_random_data
from libcloud.utils.py3 import b, httplib, parse_qs, urlparse
from libcloud.common.types import InvalidCredsError
from libcloud.storage.base import Object, Container
from libcloud.test.secrets import STORAGE_OSS_PARAMS
from libcloud.storage.types import (
    ContainerError,
    ObjectDoesNotExistError,
    ObjectHashMismatchError,
    ContainerIsNotEmptyError,
    InvalidContainerNameError,
    ContainerDoesNotExistError,
)
from libcloud.test.file_fixtures import StorageFileFixtures  # pylint: disable-msg=E0611
from libcloud.storage.drivers.oss import CHUNK_SIZE, OSSConnection, OSSStorageDriver
from libcloud.storage.drivers.dummy import DummyIterator


class OSSConnectionTestCase(unittest.TestCase):
    def setUp(self):
        self.conn = OSSConnection(
            "44CF9590006BF252F707", "OtxrzxIsfpFjA7SwPzILwy8Bw21TLhquhboDYROV"
        )

    def test_signature(self):
        expected = b("26NBxoKdsyly4EDv6inkoDft/yA=")
        headers = {
            "Content-MD5": "ODBGOERFMDMzQTczRUY3NUE3NzA5QzdFNUYzMDQxNEM=",
            "Content-Type": "text/html",
            "Expires": "Thu, 17 Nov 2005 18:49:58 GMT",
            "X-OSS-Meta-Author": "foo@bar.com",
            "X-OSS-Magic": "abracadabra",
            "Host": "oss-example.oss-cn-hangzhou.aliyuncs.com",
        }
        action = "/oss-example/nelson"
        actual = OSSConnection._get_auth_signature(
            "PUT", headers, {}, headers["Expires"], self.conn.key, action, "x-oss-"
        )
        self.assertEqual(expected, actual)


class ObjectTestCase(unittest.TestCase):
    def test_object_with_chinese_name(self):
        driver = OSSStorageDriver(*STORAGE_OSS_PARAMS)
        obj = Object(
            name="中文",
            size=0,
            hash=None,
            extra=None,
            meta_data=None,
            container=None,
            driver=driver,
        )
        self.assertTrue(obj.__repr__() is not None)


class OSSMockHttp(MockHttp, unittest.TestCase):
    fixtures = StorageFileFixtures("oss")
    base_headers = {}

    def _unauthorized(self, method, url, body, headers):
        return (
            httplib.UNAUTHORIZED,
            "",
            self.base_headers,
            httplib.responses[httplib.OK],
        )

    def _list_containers_empty(self, method, url, body, headers):
        body = self.fixtures.load("list_containers_empty.xml")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _list_containers(self, method, url, body, headers):
        body = self.fixtures.load("list_containers.xml")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _list_container_objects_empty(self, method, url, body, headers):
        body = self.fixtures.load("list_container_objects_empty.xml")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _list_container_objects(self, method, url, body, headers):
        body = self.fixtures.load("list_container_objects.xml")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _list_container_objects_chinese(self, method, url, body, headers):
        body = self.fixtures.load("list_container_objects_chinese.xml")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _list_container_objects_prefix(self, method, url, body, headers):
        params = {"prefix": self.test.prefix}
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("list_container_objects_prefix.xml")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _get_container(self, method, url, body, headers):
        return self._list_containers(method, url, body, headers)

    def _get_object(self, method, url, body, headers):
        return self._list_containers(method, url, body, headers)

    def _notexisted_get_object(self, method, url, body, headers):
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _test_get_object(self, method, url, body, headers):
        self.base_headers.update(
            {
                "accept-ranges": "bytes",
                "connection": "keep-alive",
                "content-length": "0",
                "content-type": "application/octet-stream",
                "date": "Sat, 16 Jan 2016 15:38:14 GMT",
                "etag": '"D41D8CD98F00B204E9800998ECF8427E"',
                "last-modified": "Fri, 15 Jan 2016 14:43:15 GMT",
                "server": "AliyunOSS",
                "x-oss-object-type": "Normal",
                "x-oss-request-id": "569A63E6257784731E3D877F",
                "x-oss-meta-rabbits": "monkeys",
            }
        )

        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _invalid_name(self, method, url, body, headers):
        # test_create_container_bad_request
        return (httplib.BAD_REQUEST, body, headers, httplib.responses[httplib.OK])

    def _already_exists(self, method, url, body, headers):
        # test_create_container_already_existed
        return (httplib.CONFLICT, body, headers, httplib.responses[httplib.OK])

    def _create_container(self, method, url, body, headers):
        # test_create_container_success
        self.assertEqual("PUT", method)
        self.assertEqual("", body)
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _create_container_location(self, method, url, body, headers):
        # test_create_container_success
        self.assertEqual("PUT", method)
        location_constraint = (
            "<CreateBucketConfiguration>"
            "<LocationConstraint>%s</LocationConstraint>"
            "</CreateBucketConfiguration>" % self.test.ex_location
        )
        self.assertEqual(location_constraint, body)
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _delete_container_doesnt_exist(self, method, url, body, headers):
        # test_delete_container_doesnt_exist
        return (httplib.NOT_FOUND, body, headers, httplib.responses[httplib.OK])

    def _delete_container_not_empty(self, method, url, body, headers):
        # test_delete_container_not_empty
        return (httplib.CONFLICT, body, headers, httplib.responses[httplib.OK])

    def _delete_container(self, method, url, body, headers):
        return (
            httplib.NO_CONTENT,
            body,
            self.base_headers,
            httplib.responses[httplib.NO_CONTENT],
        )

    def _foo_bar_object_not_found(self, method, url, body, headers):
        # test_delete_object_not_found
        return (httplib.NOT_FOUND, body, headers, httplib.responses[httplib.OK])

    def _foo_bar_object_delete(self, method, url, body, headers):
        # test_delete_object
        return (httplib.NO_CONTENT, body, headers, httplib.responses[httplib.OK])

    def _list_multipart(self, method, url, body, headers):
        query_string = urlparse.urlsplit(url).query
        query = parse_qs(query_string)

        if "key-marker" not in query:
            body = self.fixtures.load("ex_iterate_multipart_uploads_p1.xml")
        else:
            body = self.fixtures.load("ex_iterate_multipart_uploads_p2.xml")

        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _foo_bar_object(self, method, url, body, headers):
        # test_download_object_success
        body = generate_random_data(1000)
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _foo_bar_object_invalid_size(self, method, url, body, headers):
        # test_upload_object_invalid_file_size
        body = ""
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _foo_test_stream_data_multipart(self, method, url, body, headers):
        headers = {}
        body = ""
        headers = {"etag": '"0cc175b9c0f1b6a831c399e269772661"'}
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])


class OSSStorageDriverTestCase(unittest.TestCase):
    driver_type = OSSStorageDriver
    driver_args = STORAGE_OSS_PARAMS
    mock_response_klass = OSSMockHttp

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args)

    def setUp(self):
        self.driver_type.connectionCls.conn_class = self.mock_response_klass
        self.mock_response_klass.type = None
        self.mock_response_klass.test = self
        self.driver = self.create_driver()

    def tearDown(self):
        self._remove_test_file()

    def _remove_test_file(self):
        file_path = os.path.abspath(__file__) + ".temp"

        try:
            os.unlink(file_path)
        except OSError:
            pass

    def test_invalid_credentials(self):
        self.mock_response_klass.type = "unauthorized"
        self.assertRaises(InvalidCredsError, self.driver.list_containers)

    def test_list_containers_empty(self):
        self.mock_response_klass.type = "list_containers_empty"
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 0)

    def test_list_containers_success(self):
        self.mock_response_klass.type = "list_containers"
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 2)

        container = containers[0]
        self.assertEqual("xz02tphky6fjfiuc0", container.name)
        self.assertTrue("creation_date" in container.extra)
        self.assertEqual("2014-05-15T11:18:32.000Z", container.extra["creation_date"])
        self.assertTrue("location" in container.extra)
        self.assertEqual("oss-cn-hangzhou-a", container.extra["location"])
        self.assertEqual(self.driver, container.driver)

    def test_list_container_objects_empty(self):
        self.mock_response_klass.type = "list_container_objects_empty"
        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 0)

    def test_list_container_objects_success(self):
        self.mock_response_klass.type = "list_container_objects"
        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 2)

        obj = objects[0]
        self.assertEqual(obj.name, "en/")
        self.assertEqual(obj.hash, "D41D8CD98F00B204E9800998ECF8427E")
        self.assertEqual(obj.size, 0)
        self.assertEqual(obj.container.name, "test_container")
        self.assertEqual(obj.extra["last_modified"], "2016-01-15T14:43:15.000Z")
        self.assertTrue("owner" in obj.meta_data)

    def test_list_container_objects_with_chinese(self):
        self.mock_response_klass.type = "list_container_objects_chinese"
        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 2)

        obj = [o for o in objects if o.name == "WEB控制台.odp"][0]
        self.assertEqual(obj.hash, "281371EA1618CF0E645D6BB90A158276")
        self.assertEqual(obj.size, 1234567)
        self.assertEqual(obj.container.name, "test_container")
        self.assertEqual(obj.extra["last_modified"], "2016-01-15T14:43:06.000Z")
        self.assertTrue("owner" in obj.meta_data)

    def test_list_container_objects_with_prefix(self):
        self.mock_response_klass.type = "list_container_objects_prefix"
        container = Container(name="test_container", extra={}, driver=self.driver)
        self.prefix = "test_prefix"
        objects = self.driver.list_container_objects(container=container, prefix=self.prefix)
        self.assertEqual(len(objects), 2)

    def test_get_container_doesnt_exist(self):
        self.mock_response_klass.type = "get_container"
        self.assertRaises(
            ContainerDoesNotExistError,
            self.driver.get_container,
            container_name="not-existed",
        )

    def test_get_container_success(self):
        self.mock_response_klass.type = "get_container"
        container = self.driver.get_container(container_name="xz02tphky6fjfiuc0")
        self.assertTrue(container.name, "xz02tphky6fjfiuc0")

    def test_get_object_container_doesnt_exist(self):
        self.mock_response_klass.type = "get_object"
        self.assertRaises(
            ObjectDoesNotExistError,
            self.driver.get_object,
            container_name="xz02tphky6fjfiuc0",
            object_name="notexisted",
        )

    def test_get_object_success(self):
        self.mock_response_klass.type = "get_object"
        obj = self.driver.get_object(container_name="xz02tphky6fjfiuc0", object_name="test")

        self.assertEqual(obj.name, "test")
        self.assertEqual(obj.container.name, "xz02tphky6fjfiuc0")
        self.assertEqual(obj.size, 0)
        self.assertEqual(obj.hash, "D41D8CD98F00B204E9800998ECF8427E")
        self.assertEqual(obj.extra["last_modified"], "Fri, 15 Jan 2016 14:43:15 GMT")
        self.assertEqual(obj.extra["content_type"], "application/octet-stream")
        self.assertEqual(obj.meta_data["rabbits"], "monkeys")

    def test_create_container_bad_request(self):
        # invalid container name, returns a 400 bad request
        self.mock_response_klass.type = "invalid_name"
        self.assertRaises(
            ContainerError, self.driver.create_container, container_name="invalid_name"
        )

    def test_create_container_already_exists(self):
        # container with this name already exists
        self.mock_response_klass.type = "already_exists"
        self.assertRaises(
            InvalidContainerNameError,
            self.driver.create_container,
            container_name="new-container",
        )

    def test_create_container_success(self):
        # success
        self.mock_response_klass.type = "create_container"
        name = "new_container"
        container = self.driver.create_container(container_name=name)
        self.assertEqual(container.name, name)

    def test_create_container_with_ex_location(self):
        self.mock_response_klass.type = "create_container_location"
        name = "new_container"
        self.ex_location = "oss-cn-beijing"
        container = self.driver.create_container(container_name=name, ex_location=self.ex_location)
        self.assertEqual(container.name, name)
        self.assertTrue(container.extra["location"], self.ex_location)

    def test_delete_container_doesnt_exist(self):
        container = Container(name="new_container", extra=None, driver=self.driver)
        self.mock_response_klass.type = "delete_container_doesnt_exist"
        self.assertRaises(
            ContainerDoesNotExistError,
            self.driver.delete_container,
            container=container,
        )

    def test_delete_container_not_empty(self):
        container = Container(name="new_container", extra=None, driver=self.driver)
        self.mock_response_klass.type = "delete_container_not_empty"
        self.assertRaises(
            ContainerIsNotEmptyError, self.driver.delete_container, container=container
        )

    def test_delete_container_success(self):
        self.mock_response_klass.type = "delete_container"
        container = Container(name="new_container", extra=None, driver=self.driver)
        self.assertTrue(self.driver.delete_container(container=container))

    def test_download_object_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertTrue(result)

    def test_download_object_invalid_file_size(self):
        self.mock_response_klass.type = "invalid_size"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertFalse(result)

    def test_download_object_not_found(self):
        self.mock_response_klass.type = "not_found"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        self.assertRaises(
            ObjectDoesNotExistError,
            self.driver.download_object,
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )

    def test_download_object_as_stream_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )

        stream = self.driver.download_object_as_stream(obj=obj, chunk_size=None)
        self.assertTrue(hasattr(stream, "__iter__"))

    def test_upload_object_invalid_hash1(self):
        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
            container=None,
        ):
            return {
                "response": make_response(200, headers={"etag": "2345"}),
                "bytes_transferred": 1000,
                "data_hash": "hash343hhash89h932439jsaa89",
            }

        self.mock_response_klass.type = "INVALID_HASH1"

        old_func = self.driver_type._upload_object
        self.driver_type._upload_object = upload_file
        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        try:
            self.driver.upload_object(
                file_path=file_path,
                container=container,
                object_name=object_name,
                verify_hash=True,
            )
        except ObjectHashMismatchError:
            pass
        else:
            self.fail("Invalid hash was returned but an exception was not thrown")
        finally:
            self.driver_type._upload_object = old_func

    def test_upload_object_success(self):
        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
            container=None,
        ):
            return {
                "response": make_response(
                    200, headers={"etag": "0cc175b9c0f1b6a831c399e269772661"}
                ),
                "bytes_transferred": 1000,
                "data_hash": "0cc175b9c0f1b6a831c399e269772661",
            }

        self.mock_response_klass.type = None
        old_func = self.driver_type._upload_object
        self.driver_type._upload_object = upload_file
        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        extra = {"meta_data": {"some-value": "foobar"}}
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=True,
        )
        self.assertEqual(obj.name, "foo_test_upload")
        self.assertEqual(obj.size, 1000)
        self.assertTrue("some-value" in obj.meta_data)
        self.driver_type._upload_object = old_func

    def test_upload_object_with_acl(self):
        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
            container=None,
        ):
            return {
                "response": make_response(
                    200, headers={"etag": "0cc175b9c0f1b6a831c399e269772661"}
                ),
                "bytes_transferred": 1000,
                "data_hash": "0cc175b9c0f1b6a831c399e269772661",
            }

        self.mock_response_klass.type = None
        old_func = self.driver_type._upload_object
        self.driver_type._upload_object = upload_file

        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        extra = {"acl": "public-read"}
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=True,
        )
        self.assertEqual(obj.name, "foo_test_upload")
        self.assertEqual(obj.size, 1000)
        self.assertEqual(obj.extra["acl"], "public-read")
        self.driver_type._upload_object = old_func

    def test_upload_object_with_invalid_acl(self):
        file_path = os.path.abspath(__file__)
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        extra = {"acl": "invalid-acl"}
        self.assertRaises(
            AttributeError,
            self.driver.upload_object,
            file_path=file_path,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=True,
        )

    def test_upload_empty_object_via_stream(self):
        if self.driver.supports_multipart_upload:
            self.mock_response_klass.type = "multipart"
        else:
            self.mock_response_klass.type = None

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_stream_data"
        iterator = DummyIterator(data=[""])
        extra = {"content_type": "text/plain"}
        obj = self.driver.upload_object_via_stream(
            container=container, object_name=object_name, iterator=iterator, extra=extra
        )

        self.assertEqual(obj.name, object_name)
        self.assertEqual(obj.size, 0)

    def test_upload_small_object_via_stream(self):
        if self.driver.supports_multipart_upload:
            self.mock_response_klass.type = "multipart"
        else:
            self.mock_response_klass.type = None

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_stream_data"
        iterator = DummyIterator(data=["2", "3", "5"])
        extra = {"content_type": "text/plain"}
        obj = self.driver.upload_object_via_stream(
            container=container, object_name=object_name, iterator=iterator, extra=extra
        )

        self.assertEqual(obj.name, object_name)
        self.assertEqual(obj.size, 3)

    def test_upload_big_object_via_stream(self):
        if self.driver.supports_multipart_upload:
            self.mock_response_klass.type = "multipart"
        else:
            self.mock_response_klass.type = None

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_stream_data"
        iterator = DummyIterator(data=["2" * CHUNK_SIZE, "3" * CHUNK_SIZE, "5"])
        extra = {"content_type": "text/plain"}
        obj = self.driver.upload_object_via_stream(
            container=container, object_name=object_name, iterator=iterator, extra=extra
        )

        self.assertEqual(obj.name, object_name)
        self.assertEqual(obj.size, CHUNK_SIZE * 2 + 1)

    def test_upload_object_via_stream_abort(self):
        if not self.driver.supports_multipart_upload:
            return

        self.mock_response_klass.type = "MULTIPART"

        def _faulty_iterator():
            for i in range(0, 5):
                yield str(i)
            raise RuntimeError("Error in fetching data")

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_stream_data"
        iterator = _faulty_iterator()
        extra = {"content_type": "text/plain"}

        try:
            self.driver.upload_object_via_stream(
                container=container,
                object_name=object_name,
                iterator=iterator,
                extra=extra,
            )
        except Exception:
            pass

        return

    def test_ex_iterate_multipart_uploads(self):
        if not self.driver.supports_multipart_upload:
            return

        self.mock_response_klass.type = "list_multipart"

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        for upload in self.driver.ex_iterate_multipart_uploads(container, max_uploads=2):
            self.assertTrue(upload.key is not None)
            self.assertTrue(upload.id is not None)
            self.assertTrue(upload.initiated is not None)

    def test_ex_abort_all_multipart_uploads(self):
        if not self.driver.supports_multipart_upload:
            return

        self.mock_response_klass.type = "list_multipart"

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        with mock.patch(
            "libcloud.storage.drivers.oss.OSSStorageDriver" "._abort_multipart",
            autospec=True,
        ) as mock_abort:
            self.driver.ex_abort_all_multipart_uploads(container)

            self.assertEqual(3, mock_abort.call_count)

    def test_delete_object_not_found(self):
        self.mock_response_klass.type = "not_found"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1234,
            hash=None,
            extra=None,
            meta_data=None,
            container=container,
            driver=self.driver,
        )
        self.assertRaises(ObjectDoesNotExistError, self.driver.delete_object, obj=obj)

    def test_delete_object_success(self):
        self.mock_response_klass.type = "delete"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1234,
            hash=None,
            extra=None,
            meta_data=None,
            container=container,
            driver=self.driver,
        )

        result = self.driver.delete_object(obj=obj)
        self.assertTrue(result)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import json
import tempfile
from io import BytesIO

from libcloud.test import generate_random_data  # pylint: disable-msg=E0611
from libcloud.test import unittest
from libcloud.utils.py3 import b, httplib, parse_qs, urlparse, basestring
from libcloud.common.types import LibcloudError, InvalidCredsError
from libcloud.storage.base import Object, Container
from libcloud.test.secrets import STORAGE_AZURE_BLOBS_PARAMS, STORAGE_AZURITE_BLOBS_PARAMS
from libcloud.storage.types import (
    ObjectDoesNotExistError,
    ObjectHashMismatchError,
    ContainerIsNotEmptyError,
    InvalidContainerNameError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)
from libcloud.test.storage.base import BaseRangeDownloadMockHttp
from libcloud.test.file_fixtures import StorageFileFixtures  # pylint: disable-msg=E0611
from libcloud.storage.drivers.azure_blobs import (
    AZURE_UPLOAD_CHUNK_SIZE,
    AzureBlobsStorageDriver,
    AzureBlobsActiveDirectoryConnection,
)


class AzureBlobsMockHttp(BaseRangeDownloadMockHttp, unittest.TestCase):
    fixtures = StorageFileFixtures("azure_blobs")
    base_headers = {}

    # Note: using this method to get the oauth key for azure ad authentication
    def __getattr__(self, n):
        def fn(method, url, body, headers):
            fixture = self.fixtures.load(n + ".json")

            if method in ("POST", "PUT"):
                try:
                    body = json.loads(body)
                    fixture_tmp = json.loads(fixture)
                    fixture_tmp = self._update(fixture_tmp, body)
                    fixture = json.dumps(fixture_tmp)
                except ValueError:
                    pass

            return (httplib.OK, fixture, headers, httplib.responses[httplib.OK])

        return fn

    def _UNAUTHORIZED(self, method, url, body, headers):
        return (
            httplib.UNAUTHORIZED,
            "",
            self.base_headers,
            httplib.responses[httplib.UNAUTHORIZED],
        )

    def _list_containers_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("list_containers_empty.xml")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _list_containers(self, method, url, body, headers):
        query_string = urlparse.urlsplit(url).query
        query = parse_qs(query_string)

        if "marker" not in query:
            body = self.fixtures.load("list_containers_1.xml")
        else:
            body = self.fixtures.load("list_containers_2.xml")

        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _test_container_EMPTY(self, method, url, body, headers):
        if method == "DELETE":
            body = ""
            return (
                httplib.ACCEPTED,
                body,
                self.base_headers,
                httplib.responses[httplib.ACCEPTED],
            )

        else:
            body = self.fixtures.load("list_objects_empty.xml")
            return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _new__container_INVALID_NAME(self, method, url, body, headers):
        return (
            httplib.BAD_REQUEST,
            body,
            self.base_headers,
            httplib.responses[httplib.BAD_REQUEST],
        )

    def _test_container(self, method, url, body, headers):
        query_string = urlparse.urlsplit(url).query
        query = parse_qs(query_string)

        if "marker" not in query:
            body = self.fixtures.load("list_objects_1.xml")
        else:
            body = self.fixtures.load("list_objects_2.xml")

        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _test_container100(self, method, url, body, headers):
        body = ""

        if method != "HEAD":
            return (
                httplib.BAD_REQUEST,
                body,
                self.base_headers,
                httplib.responses[httplib.BAD_REQUEST],
            )

        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _test_container200(self, method, url, body, headers):
        body = ""

        if method != "HEAD":
            return (
                httplib.BAD_REQUEST,
                body,
                self.base_headers,
                httplib.responses[httplib.BAD_REQUEST],
            )

        headers = {}

        headers["etag"] = "0x8CFB877BB56A6FB"
        headers["last-modified"] = "Fri, 04 Jan 2013 09:48:06 GMT"
        headers["x-ms-lease-status"] = "unlocked"
        headers["x-ms-lease-state"] = "available"
        headers["x-ms-meta-meta1"] = "value1"

        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _test_container200_test(self, method, url, body, headers):
        body = ""

        if method != "HEAD":
            return (
                httplib.BAD_REQUEST,
                body,
                self.base_headers,
                httplib.responses[httplib.BAD_REQUEST],
            )

        headers = {}

        headers["etag"] = "0x8CFB877BB56A6FB"
        headers["last-modified"] = "Fri, 04 Jan 2013 09:48:06 GMT"
        headers["content-length"] = "12345"
        headers["content-type"] = "application/zip"
        headers["x-ms-blob-type"] = "Block"
        headers["x-ms-lease-status"] = "unlocked"
        headers["x-ms-lease-state"] = "available"
        headers["x-ms-meta-rabbits"] = "monkeys"

        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _test2_test_list_containers(self, method, url, body, headers):
        # test_get_object
        body = self.fixtures.load("list_containers.xml")
        headers = {
            "content-type": "application/zip",
            "etag": '"e31208wqsdoj329jd"',
            "x-amz-meta-rabbits": "monkeys",
            "content-length": "12345",
            "last-modified": "Thu, 13 Sep 2012 07:13:22 GMT",
        }

        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _new_container_ALREADY_EXISTS(self, method, url, body, headers):
        # test_create_container
        return (httplib.CONFLICT, body, headers, httplib.responses[httplib.CONFLICT])

    def _new_container(self, method, url, body, headers):
        # test_create_container, test_delete_container

        headers = {}

        if method == "PUT":
            status = httplib.CREATED

            headers["etag"] = "0x8CFB877BB56A6FB"
            headers["last-modified"] = "Fri, 04 Jan 2013 09:48:06 GMT"
            headers["x-ms-lease-status"] = "unlocked"
            headers["x-ms-lease-state"] = "available"
            headers["x-ms-meta-meta1"] = "value1"

        elif method == "DELETE":
            status = httplib.NO_CONTENT

        return (status, body, headers, httplib.responses[status])

    def _new_container_DOESNT_EXIST(self, method, url, body, headers):
        # test_delete_container
        return (httplib.NOT_FOUND, body, headers, httplib.responses[httplib.NOT_FOUND])

    def _foo_bar_container_NOT_FOUND(self, method, url, body, headers):
        # test_delete_container_not_found
        return (httplib.NOT_FOUND, body, headers, httplib.responses[httplib.NOT_FOUND])

    def _foo_bar_container_foo_bar_object_NOT_FOUND(self, method, url, body, headers):
        # test_delete_object_not_found
        return (httplib.NOT_FOUND, body, headers, httplib.responses[httplib.NOT_FOUND])

    def _foo_bar_container_foo_bar_object_DELETE(self, method, url, body, headers):
        # test_delete_object
        return (httplib.ACCEPTED, body, headers, httplib.responses[httplib.ACCEPTED])

    def _foo_bar_container_foo_test_upload(self, method, url, body, headers):
        self._assert_content_length_header_is_string(headers=headers)

        query_string = urlparse.urlsplit(url).query
        query = parse_qs(query_string)
        comp = query.get("comp", [])

        headers = {}
        body = ""

        if "blocklist" in comp or not comp:
            headers["etag"] = '"0x8CFB877BB56A6FB"'
            headers["content-md5"] = "d4fe4c9829f7ca1cc89db7ad670d2bbd"
        elif "block" in comp:
            headers["content-md5"] = "lvcfx/bOJvndpRlrdKU1YQ=="
        else:
            raise NotImplementedError("Unknown request comp: {}".format(comp))

        return (httplib.CREATED, body, headers, httplib.responses[httplib.CREATED])

    def _foo_bar_container_foo_test_upload_block(self, method, url, body, headers):
        # test_upload_object_success
        self._assert_content_length_header_is_string(headers=headers)

        body = ""
        headers = {}
        headers["etag"] = "0x8CFB877BB56A6FB"
        return (httplib.CREATED, body, headers, httplib.responses[httplib.CREATED])

    def _foo_bar_container_foo_test_upload_blocklist(self, method, url, body, headers):
        # test_upload_object_success
        self._assert_content_length_header_is_string(headers=headers)

        body = ""
        headers = {}
        headers["etag"] = "0x8CFB877BB56A6FB"
        headers["content-md5"] = "d4fe4c9829f7ca1cc89db7ad670d2bbd"

        return (httplib.CREATED, body, headers, httplib.responses[httplib.CREATED])

    def _foo_bar_container_foo_test_upload_lease(self, method, url, body, headers):
        # test_upload_object_success
        self._assert_content_length_header_is_string(headers=headers)

        action = headers["x-ms-lease-action"]
        rheaders = {"x-ms-lease-id": "someleaseid"}
        body = ""

        if action == "acquire":
            return (httplib.CREATED, body, rheaders, httplib.responses[httplib.CREATED])

        else:
            if headers.get("x-ms-lease-id", None) != "someleaseid":
                return (
                    httplib.BAD_REQUEST,
                    body,
                    rheaders,
                    httplib.responses[httplib.BAD_REQUEST],
                )

            return (httplib.OK, body, headers, httplib.responses[httplib.CREATED])

    def _foo_bar_container_foo_test_upload_INVALID_HASH(self, method, url, body, headers):
        # test_upload_object_invalid_hash1
        self._assert_content_length_header_is_string(headers=headers)

        body = ""
        headers = {}
        headers["etag"] = "0x8CFB877BB56A6FB"
        headers["content-md5"] = "d4fe4c9829f7ca1cc89db7ad670d2bbd"

        return (httplib.CREATED, body, headers, httplib.responses[httplib.CREATED])

    def _foo_bar_container_foo_bar_object(self, method, url, body, headers):
        # test_upload_object_invalid_file_size
        self._assert_content_length_header_is_string(headers=headers)

        body = generate_random_data(1000)
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _foo_bar_container_foo_bar_object_range(self, method, url, body, headers):
        # test_download_object_range_success
        body = "0123456789123456789"

        self.assertTrue("x-ms-range" in headers)
        self.assertEqual(headers["x-ms-range"], "bytes=5-6")

        start_bytes, end_bytes = self._get_start_and_end_bytes_from_range_str(
            headers["x-ms-range"], body
        )

        return (
            httplib.PARTIAL_CONTENT,
            body[start_bytes : end_bytes + 1],
            headers,
            httplib.responses[httplib.PARTIAL_CONTENT],
        )

    def _foo_bar_container_foo_bar_object_range_stream(self, method, url, body, headers):
        # test_download_object_range_as_stream_success
        body = "0123456789123456789"

        self.assertTrue("x-ms-range" in headers)
        self.assertEqual(headers["x-ms-range"], "bytes=4-5")

        start_bytes, end_bytes = self._get_start_and_end_bytes_from_range_str(
            headers["x-ms-range"], body
        )

        return (
            httplib.PARTIAL_CONTENT,
            body[start_bytes : end_bytes + 1],
            headers,
            httplib.responses[httplib.PARTIAL_CONTENT],
        )

    def _foo_bar_container_foo_bar_object_INVALID_SIZE(self, method, url, body, headers):
        # test_upload_object_invalid_file_size
        self._assert_content_length_header_is_string(headers=headers)

        body = ""
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _assert_content_length_header_is_string(self, headers):
        if "Content-Length" in headers:
            self.assertTrue(isinstance(headers["Content-Length"], basestring))


class AzuriteBlobsMockHttp(AzureBlobsMockHttp):
    fixtures = StorageFileFixtures("azurite_blobs")

    def _get_method_name(self, *args, **kwargs):
        method_name = super()._get_method_name(*args, **kwargs)

        if method_name.startswith("_account"):
            method_name = method_name[8:]

        return method_name


class AzureBlobsTests(unittest.TestCase):
    driver_type = AzureBlobsStorageDriver
    driver_args = STORAGE_AZURE_BLOBS_PARAMS
    mock_response_klass = AzureBlobsMockHttp

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args)

    def setUp(self):
        self.driver_type.connectionCls.conn_class = self.mock_response_klass
        self.mock_response_klass.type = None
        self.driver = self.create_driver()

    def tearDown(self):
        self._remove_test_file()

    def _remove_test_file(self):
        file_path = os.path.abspath(__file__) + ".temp"

        try:
            os.unlink(file_path)
        except OSError:
            pass

    def test_invalid_credentials(self):
        self.mock_response_klass.type = "UNAUTHORIZED"
        try:
            self.driver.list_containers()
        except InvalidCredsError as e:
            self.assertEqual(True, isinstance(e, InvalidCredsError))
        else:
            self.fail("Exception was not thrown")

    def test_list_containers_empty(self):
        self.mock_response_klass.type = "list_containers_EMPTY"
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 0)

    def test_list_containers_success(self):
        self.mock_response_klass.type = "list_containers"
        AzureBlobsStorageDriver.RESPONSES_PER_REQUEST = 2
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 4)

        self.assertTrue("last_modified" in containers[1].extra)
        self.assertTrue("url" in containers[1].extra)
        self.assertTrue("etag" in containers[1].extra)
        self.assertTrue("lease" in containers[1].extra)
        self.assertTrue("meta_data" in containers[1].extra)
        self.assertEqual(containers[1].extra["etag"], "0x8CFBAB7B5B82D8E")

    def test_list_container_objects_empty(self):
        self.mock_response_klass.type = "EMPTY"
        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 0)

    def test_list_container_objects_success(self):
        self.mock_response_klass.type = None
        AzureBlobsStorageDriver.RESPONSES_PER_REQUEST = 2

        container = Container(name="test_container", extra={}, driver=self.driver)

        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 4)

        obj = objects[1]
        self.assertEqual(obj.name, "object2.txt")
        self.assertEqual(obj.hash, "0x8CFB90F1BA8CD8F")
        self.assertEqual(obj.size, 1048576)
        self.assertEqual(obj.container.name, "test_container")
        self.assertTrue("meta1" in obj.meta_data)
        self.assertTrue("meta2" in obj.meta_data)
        self.assertTrue("last_modified" in obj.extra)
        self.assertTrue("content_type" in obj.extra)
        self.assertTrue("content_encoding" in obj.extra)
        self.assertTrue("content_language" in obj.extra)

    def test_list_container_objects_with_prefix(self):
        self.mock_response_klass.type = None
        AzureBlobsStorageDriver.RESPONSES_PER_REQUEST = 2

        container = Container(name="test_container", extra={}, driver=self.driver)
        objects = self.driver.list_container_objects(container=container, prefix="test_prefix")
        self.assertEqual(len(objects), 4)

        obj = objects[1]
        self.assertEqual(obj.name, "object2.txt")
        self.assertEqual(obj.hash, "0x8CFB90F1BA8CD8F")
        self.assertEqual(obj.size, 1048576)
        self.assertEqual(obj.container.name, "test_container")
        self.assertTrue("meta1" in obj.meta_data)
        self.assertTrue("meta2" in obj.meta_data)
        self.assertTrue("last_modified" in obj.extra)
        self.assertTrue("content_type" in obj.extra)
        self.assertTrue("content_encoding" in obj.extra)
        self.assertTrue("content_language" in obj.extra)

    def test_get_container_doesnt_exist(self):
        self.mock_response_klass.type = None
        try:
            self.driver.get_container(container_name="test_container100")
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_container_success(self):
        self.mock_response_klass.type = None
        container = self.driver.get_container(container_name="test_container200")

        self.assertTrue(container.name, "test_container200")
        self.assertTrue(container.extra["etag"], "0x8CFB877BB56A6FB")
        self.assertTrue(container.extra["last_modified"], "Fri, 04 Jan 2013 09:48:06 GMT")
        self.assertTrue(container.extra["lease"]["status"], "unlocked")
        self.assertTrue(container.extra["lease"]["state"], "available")
        self.assertTrue(container.extra["meta_data"]["meta1"], "value1")

        if self.driver.secure:
            expected_url = "https://account.blob.core.windows.net/test_container200"
        else:
            expected_url = "http://localhost/account/test_container200"

        self.assertEqual(container.extra["url"], expected_url)

    def test_get_object_cdn_url(self):
        obj = self.driver.get_object(container_name="test_container200", object_name="test")

        url = urlparse.urlparse(self.driver.get_object_cdn_url(obj))
        query = urlparse.parse_qs(url.query)

        self.assertEqual(len(query["sig"]), 1)
        self.assertGreater(len(query["sig"][0]), 0)

    def test_get_object_container_doesnt_exist(self):
        # This method makes two requests which makes mocking the response a bit
        # trickier
        self.mock_response_klass.type = None
        try:
            self.driver.get_object(container_name="test_container100", object_name="test")
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_object_success(self):
        # This method makes two requests which makes mocking the response a bit
        # trickier
        self.mock_response_klass.type = None
        obj = self.driver.get_object(container_name="test_container200", object_name="test")

        self.assertEqual(obj.name, "test")
        self.assertEqual(obj.container.name, "test_container200")
        self.assertEqual(obj.size, 12345)
        self.assertEqual(obj.hash, "0x8CFB877BB56A6FB")
        self.assertEqual(obj.extra["last_modified"], "Fri, 04 Jan 2013 09:48:06 GMT")
        self.assertEqual(obj.extra["content_type"], "application/zip")
        self.assertEqual(obj.meta_data["rabbits"], "monkeys")

    def test_create_container_invalid_name(self):
        # invalid container name
        self.mock_response_klass.type = "INVALID_NAME"
        try:
            self.driver.create_container(container_name="new--container")
        except InvalidContainerNameError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_container_already_exists(self):
        # container with this name already exists
        self.mock_response_klass.type = "ALREADY_EXISTS"
        try:
            self.driver.create_container(container_name="new-container")
        except ContainerAlreadyExistsError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_container_success(self):
        # success
        self.mock_response_klass.type = None
        name = "new-container"
        container = self.driver.create_container(container_name=name)
        self.assertEqual(container.name, name)

    def test_delete_container_doesnt_exist(self):
        container = Container(name="new_container", extra=None, driver=self.driver)
        self.mock_response_klass.type = "DOESNT_EXIST"
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_container_not_empty(self):
        self.mock_response_klass.type = None
        AzureBlobsStorageDriver.RESPONSES_PER_REQUEST = 2

        container = Container(name="test_container", extra={}, driver=self.driver)

        try:
            self.driver.delete_container(container=container)
        except ContainerIsNotEmptyError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_container_success(self):
        self.mock_response_klass.type = "EMPTY"
        AzureBlobsStorageDriver.RESPONSES_PER_REQUEST = 2

        container = Container(name="test_container", extra={}, driver=self.driver)

        self.assertTrue(self.driver.delete_container(container=container))

    def test_delete_container_not_found(self):
        self.mock_response_klass.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Container does not exist but an exception was not" + "thrown")

    def test_download_object_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertTrue(result)

    def test_download_object_invalid_file_size(self):
        self.mock_response_klass.type = "INVALID_SIZE"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertFalse(result)

    def test_download_object_invalid_file_already_exists(self):
        self.mock_response_klass.type = "INVALID_SIZE"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = os.path.abspath(__file__)
        try:
            self.driver.download_object(
                obj=obj,
                destination_path=destination_path,
                overwrite_existing=False,
                delete_on_failure=True,
            )
        except LibcloudError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_download_object_as_stream_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )

        stream = self.driver.download_object_as_stream(obj=obj, chunk_size=None)

        consumed_stream = "".join(chunk.decode("utf-8") for chunk in stream)
        self.assertEqual(len(consumed_stream), obj.size)

    def test_download_object_range_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object_range",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object_range(
            obj=obj,
            start_bytes=5,
            end_bytes=7,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertTrue(result)

        with open(destination_path) as fp:
            content = fp.read()

        self.assertEqual(content, "56")

    def test_download_object_range_as_stream_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        obj = Object(
            name="foo_bar_object_range_stream",
            size=2,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )

        stream = self.driver.download_object_range_as_stream(
            obj=obj, start_bytes=4, end_bytes=6, chunk_size=None
        )

        consumed_stream = "".join(chunk.decode("utf-8") for chunk in stream)
        self.assertEqual(consumed_stream, "45")
        self.assertEqual(len(consumed_stream), obj.size)

    def test_upload_object_invalid_md5(self):
        # Invalid md5 is returned by azure
        self.mock_response_klass.type = "INVALID_HASH"

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        file_path = os.path.abspath(__file__)
        try:
            self.driver.upload_object(
                file_path=file_path,
                container=container,
                object_name=object_name,
                verify_hash=True,
            )
        except ObjectHashMismatchError:
            pass
        else:
            self.fail("Invalid hash was returned but an exception was not thrown")

    def test_upload_small_block_object_success(self):
        file_path = os.path.abspath(__file__)
        file_size = os.stat(file_path).st_size

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        extra = {"meta_data": {"some-value": "foobar"}}
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=False,
        )

        self.assertEqual(obj.name, "foo_test_upload")
        self.assertEqual(obj.size, file_size)
        self.assertTrue("some-value" in obj.meta_data)

    def test_upload_big_block_object_success(self):
        _, file_path = tempfile.mkstemp(suffix=".jpg")
        file_size = AZURE_UPLOAD_CHUNK_SIZE + 1

        with open(file_path, "w") as file_hdl:
            file_hdl.write("0" * file_size)

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        extra = {"meta_data": {"some-value": "foobar"}}
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=False,
        )

        self.assertEqual(obj.name, "foo_test_upload")
        self.assertEqual(obj.size, file_size)
        self.assertTrue("some-value" in obj.meta_data)

        os.remove(file_path)

    def test_upload_small_block_object_success_with_lease(self):
        self.mock_response_klass.use_param = "comp"
        file_path = os.path.abspath(__file__)
        file_size = os.stat(file_path).st_size

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        extra = {"meta_data": {"some-value": "foobar"}}
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=False,
            ex_use_lease=True,
        )

        self.assertEqual(obj.name, "foo_test_upload")
        self.assertEqual(obj.size, file_size)
        self.assertTrue("some-value" in obj.meta_data)
        self.mock_response_klass.use_param = None

    def test_upload_big_block_object_success_with_lease(self):
        self.mock_response_klass.use_param = "comp"
        _, file_path = tempfile.mkstemp(suffix=".jpg")
        file_size = AZURE_UPLOAD_CHUNK_SIZE * 2

        with open(file_path, "w") as file_hdl:
            file_hdl.write("0" * file_size)

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        object_name = "foo_test_upload"
        extra = {"meta_data": {"some-value": "foobar"}}
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=False,
            ex_use_lease=False,
        )

        self.assertEqual(obj.name, "foo_test_upload")
        self.assertEqual(obj.size, file_size)
        self.assertTrue("some-value" in obj.meta_data)

        os.remove(file_path)
        self.mock_response_klass.use_param = None

    def test_upload_blob_object_via_stream(self):
        self.mock_response_klass.use_param = "comp"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        object_name = "foo_test_upload"
        iterator = BytesIO(b("345"))
        extra = {"content_type": "text/plain"}
        obj = self.driver.upload_object_via_stream(
            container=container, object_name=object_name, iterator=iterator, extra=extra
        )

        self.assertEqual(obj.name, object_name)
        self.assertEqual(obj.size, 3)
        self.mock_response_klass.use_param = None

    def test_upload_blob_object_via_stream_from_iterable(self):
        self.mock_response_klass.use_param = "comp"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        object_name = "foo_test_upload"
        iterator = iter([b("34"), b("5")])
        extra = {"content_type": "text/plain"}
        obj = self.driver.upload_object_via_stream(
            container=container, object_name=object_name, iterator=iterator, extra=extra
        )

        self.assertEqual(obj.name, object_name)
        self.assertEqual(obj.size, 3)
        self.mock_response_klass.use_param = None

    def test_upload_blob_object_via_stream_with_lease(self):
        self.mock_response_klass.use_param = "comp"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        object_name = "foo_test_upload"
        iterator = BytesIO(b("345"))
        extra = {"content_type": "text/plain"}
        obj = self.driver.upload_object_via_stream(
            container=container,
            object_name=object_name,
            iterator=iterator,
            extra=extra,
            ex_use_lease=True,
        )

        self.assertEqual(obj.name, object_name)
        self.assertEqual(obj.size, 3)
        self.mock_response_klass.use_param = None

    def test_delete_object_not_found(self):
        self.mock_response_klass.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1234,
            hash=None,
            extra=None,
            meta_data=None,
            container=container,
            driver=self.driver,
        )
        try:
            self.driver.delete_object(obj=obj)
        except ObjectDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_object_success(self):
        self.mock_response_klass.type = "DELETE"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1234,
            hash=None,
            extra=None,
            meta_data=None,
            container=container,
            driver=self.driver,
        )

        result = self.driver.delete_object(obj=obj)
        self.assertTrue(result)

    def test_storage_driver_host(self):
        # Non regression tests for issue LIBCLOUD-399 dealing with the bad
        # management of the connectionCls.host class attribute
        driver1 = self.driver_type("fakeaccount1", "deadbeafcafebabe==")
        driver2 = self.driver_type("fakeaccount2", "deadbeafcafebabe==")
        driver3 = self.driver_type("fakeaccount3", "deadbeafcafebabe==", host="test.foo.bar.com")

        host1 = driver1.connection.host
        host2 = driver2.connection.host
        host3 = driver3.connection.host

        self.assertEqual(host1, "fakeaccount1.blob.core.windows.net")
        self.assertEqual(host2, "fakeaccount2.blob.core.windows.net")
        self.assertEqual(host3, "test.foo.bar.com")

    def test_normalize_http_headers(self):
        driver = self.driver_type("fakeaccount1", "deadbeafcafebabe==")

        headers = driver._fix_headers(
            {
                # should be normalized to include x-ms-blob prefix
                "Content-Encoding": "gzip",
                "content-language": "en-us",
                # should be passed through
                "x-foo": "bar",
            }
        )

        self.assertEqual(
            headers,
            {
                "x-ms-blob-content-encoding": "gzip",
                "x-ms-blob-content-language": "en-us",
                "x-foo": "bar",
            },
        )

    def test_storage_driver_host_govcloud(self):
        driver1 = self.driver_type(
            "fakeaccount1", "deadbeafcafebabe==", host="blob.core.usgovcloudapi.net"
        )
        driver2 = self.driver_type(
            "fakeaccount2",
            "deadbeafcafebabe==",
            host="fakeaccount2.blob.core.usgovcloudapi.net",
        )

        host1 = driver1.connection.host
        host2 = driver2.connection.host
        account_prefix_1 = driver1.connection.account_prefix
        account_prefix_2 = driver2.connection.account_prefix

        self.assertEqual(host1, "fakeaccount1.blob.core.usgovcloudapi.net")
        self.assertEqual(host2, "fakeaccount2.blob.core.usgovcloudapi.net")
        self.assertIsNone(account_prefix_1)
        self.assertIsNone(account_prefix_2)

    def test_storage_driver_host_azurite(self):
        driver = self.driver_type(
            "fakeaccount1",
            "deadbeafcafebabe==",
            host="localhost",
            port=10000,
            secure=False,
        )

        host = driver.connection.host
        account_prefix = driver.connection.account_prefix

        self.assertEqual(host, "localhost")
        self.assertEqual(account_prefix, "fakeaccount1")

    def test_storage_driver_azure_ad(self):
        AzureBlobsActiveDirectoryConnection.conn_class = AzureBlobsMockHttp
        driver = self.driver_type(
            key="fakeaccount1",
            secret="DEKjfhdakkdjfhei~",
            tenant_id="77777777-7777-7777-7777-777777777777",
            identity="55555555-5555-5555-5555-555555555555",
            auth_type="azureAd",
            secure=True,
        )
        host = driver.connection.host

        self.assertEqual(host, "fakeaccount1.blob.core.windows.net")

    def test_get_azure_ad_object_success(self):
        AzureBlobsActiveDirectoryConnection.conn_class = AzureBlobsMockHttp
        driver = self.driver_type(
            key="fakeaccount1",
            secret="DEKjfhdakkdjfhei~",
            tenant_id="77777777-7777-7777-7777-777777777777",
            identity="55555555-5555-5555-5555-555555555555",
            auth_type="azureAd",
            secure=True,
        )
        self.mock_response_klass.type = None
        container = driver.get_container(container_name="test_container200")

        self.assertTrue(container.name, "test_container200")
        self.assertTrue(container.extra["etag"], "0x8CFB877BB56A6FB")
        self.assertTrue(container.extra["last_modified"], "Fri, 04 Jan 2013 09:48:06 GMT")
        self.assertTrue(container.extra["lease"]["status"], "unlocked")
        self.assertTrue(container.extra["lease"]["state"], "available")
        self.assertTrue(container.extra["meta_data"]["meta1"], "value1")


class AzuriteBlobsTests(AzureBlobsTests):
    driver_args = STORAGE_AZURITE_BLOBS_PARAMS
    mock_response_klass = AzuriteBlobsMockHttp


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from typing import Dict, List

from libcloud.utils.py3 import b
from libcloud.common.base import JsonResponse, ConnectionUserAndKey

__all__ = ["API_HOST", "LuadnsException", "LuadnsResponse", "LuadnsConnection"]

# Endpoint for luadns api
API_HOST = "api.luadns.com"


class LuadnsResponse(JsonResponse):
    errors = []  # type: List[Dict]
    objects = []  # type: List[Dict]

    def __init__(self, response, connection):
        super().__init__(response=response, connection=connection)
        self.errors, self.objects = self.parse_body_and_errors()
        if not self.success():
            raise LuadnsException(code=self.status, message=self.errors.pop()["message"])

    def parse_body_and_errors(self):
        js = super().parse_body()
        if "message" in js:
            self.errors.append(js)
        else:
            self.objects.append(js)

        return self.errors, self.objects

    def success(self):
        return len(self.errors) == 0


class LuadnsConnection(ConnectionUserAndKey):
    host = API_HOST
    responseCls = LuadnsResponse

    def add_default_headers(self, headers):
        b64string = b("{}:{}".format(self.user_id, self.key))
        encoded = base64.b64encode(b64string).decode("utf-8")
        authorization = "Basic " + encoded

        headers["Accept"] = "application/json"
        headers["Authorization"] = authorization

        return headers


class LuadnsException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return "{} {}".format(self.code, self.message)

    def __repr__(self):
        return "Luadns {} {}".format(self.code, self.message)
